from lark import Token

from ... import exc
from ..presets import reserved_keywords


def get_identifier(node, throw_exc=True) -> str:
    """takes some node from the tree, either a token or a subtree, and finds
    the identifier it contains. this is used for the rule that matches an
    identifier or a quoted identifier, because, in that case, an identifier will
    be a token, and a quoted identifier will be a subtree containing an
    identifier token.

    also we check to make sure that the identifier is not a reserved keyword, as
    a convenience. we could do it a separate step, but it's easier to do it
    here."""

    def match_ident(v):
        return isinstance(v, Token) and v.type == "IDENTIFIER"

    quoted = False
    ident = next(node.scan_values(match_ident)).value
    quoted = bool(list(node.find_data("quoted_identifier")))
    if throw_exc and not quoted and ident.lower() in reserved_keywords:
        raise exc.ReservedKeyword(keyword=ident)
    return ident



from abc import abstractmethod
from itertools import chain
from typing import Optional, Sequence, cast

from lark import Lark, ParseTree
from lark.exceptions import VisitError
from lark.reconstruct import Reconstructor

from heimdallm.bifrosts.sql import exc
from heimdallm.constraints import ConstraintValidator as _ConstraintValidator

from ...utils import ANY_JOIN, FqColumn, JoinCondition, RequiredConstraint
from .. import presets
from ..utils.visitors import AliasCollector
from .visitors import FacetCollector, Facets


class SQLConstraintValidator(_ConstraintValidator):
    """
    Validate different aspects of a SQL query. You are intended to derive this class and
    implement methods."""

    @abstractmethod
    def requester_identities(self) -> Sequence[RequiredConstraint]:
        """Returns the identity of the requester, as represented in the
        database. this is used to instruct the LLM how to constrain the query
        that it generates. only one of these identities needs to match.

        the reason that we return a sequence, and not a single identity, is that
        sometimes an LLM will specify the constraint as part of a ``JOIN``
        condition, and not a ``WHERE`` condition. in that case, the column in the
        JOIN condition may not match the column you expect, for example,

            Invoice.CustomerId vs Customer.CustomerId

        although they represent the same identity, they are different tables and
        columns, but constraining on one of them is sufficient.
        """
        raise NotImplementedError(
            "You must explicitly provide the requester identity, "
            "or an empty list for full access (dangerous)"
        )

    @abstractmethod
    def required_constraints(self) -> Sequence[RequiredConstraint]:
        """Returns a sequence of secure constraints that must exist in the ``WHERE``
        clause of the query"""
        raise NotImplementedError(
            "You must explicitly provide a sequence of required constraints, "
            "or an empty list for no constraints"
        )

    @abstractmethod
    def select_column_allowed(self, column: FqColumn) -> bool:
        """Ensures that the column is allowed to be selected in the ``SELECT`` clause"""
        return False

    @abstractmethod
    def allowed_joins(self) -> Sequence[JoinCondition]:
        """Returns all of the tables allowed to be connected to the query,
        either via a ``JOIN``, or via a ``FROM``. this encompasses both cases because
        LLMs frequently produce queries that select unpredictable tables with
        ``FROM``, even if that table is valid to be connected via ``JOIN``. so to
        handle this, we consider a table selected with ``FROM`` as a ``JOIN`` with no
        explicit conditions. in practice, if there are other joins in the query,
        the selected table will have an explicit condition via the ON clause. if
        there are no other joins, the selected table will have no join
        conditions.
        """
        return []

    @abstractmethod
    def max_limit(self) -> Optional[int]:
        """Return the maximum number of rows that can be returned by a query. if
        None, there is no limit."""
        return None

    def can_use_function(self, function: str) -> bool:
        """Ensures that the function is allowed to be used in the query

        :param function: the name of the function
        """
        return function in presets.safe_functions

    def condition_column_allowed(self, fq_column: FqColumn) -> bool:
        """Checks if a column is allowed in a ``WHERE``, ``JOIN``, ``HAVING``, or
        ``ORDER BY``"""
        # let's default to "if you can see it, you can use it"
        return self.select_column_allowed(fq_column)

    def fix(self, grammar: Lark, tree: ParseTree) -> str:
        """A parse tree may be valid SQL, but it may not be valid according to
        the validator's constraints. we may be able to make intelligent
        decisions about those constraints, and fix the parse tree though, for
        example, by adding a limit to a query."""

        # gets around a circular import issue
        from heimdallm.bifrosts.sql.sqlite.select import reconstruct

        transform = reconstruct.ReconstructTransformer(self)
        try:
            fixed_tree = transform.transform(tree)
        except VisitError as e:
            if isinstance(e.orig_exc, exc.BaseException):
                raise e.orig_exc
            raise e

        output = Reconstructor(grammar).reconstruct(
            fixed_tree,
            postproc=reconstruct.postproc,
        )
        return output

    def validate(self, untrusted_input: str, tree: ParseTree):
        """Analyze the tree and validate it against our SQL constraints"""
        try:
            alias_collector = AliasCollector()
            alias_collector.visit(tree)

            facets = Facets()
            facet_collector = FacetCollector(facets, alias_collector)
            facet_collector.visit(tree)
        except exc.GeneralParseError:
            raise exc.InvalidQuery(query=untrusted_input)

        # check the select column allowlist
        for fq_column in facets.selected_columns:
            if not self.select_column_allowed(fq_column):
                raise exc.IllegalSelectedColumn(column=fq_column.name)

        # check the join condition allowlist
        any_join_cond_allowed = ANY_JOIN in self.allowed_joins()
        for table, join_specs in facets.joined_tables.items():
            for join_spec in join_specs:
                allowed = any_join_cond_allowed or join_spec in self.allowed_joins()
                if not allowed:
                    raise exc.IllegalJoinTable(join=join_spec)

        # ensure that the selected table is joined to another table, if joins exist
        if (
            facets.joined_tables
            and not facets.joined_tables[cast(str, facets.selected_table)]
        ):
            raise exc.DisconnectedTable(table=facets.selected_table)

        # ensure that all joins were joined on columns that exist on the joined tables
        if facets.bad_joins:
            raise exc.BogusJoinedTable(
                table=facets.bad_joins[0],
            )

        # all columns specified in the join conditions are automatically included in the
        # condition column allowlist
        allowed_join_conditions = set()
        for join_spec in self.allowed_joins():
            if join_spec == ANY_JOIN:
                continue
            allowed_join_conditions.add(join_spec.first)
            allowed_join_conditions.add(join_spec.second)

        # check the condition column allowlist
        for fq_column in facets.condition_columns:
            if fq_column in allowed_join_conditions:
                continue

            if not self.condition_column_allowed(fq_column):
                raise exc.IllegalConditionColumn(
                    column=fq_column,
                )

        # get all of the constraints that the query MUST be constrained by in
        # the WHERE clause
        for constraint in self.required_constraints():
            if constraint not in facets.required_constraints:
                raise exc.MissingRequiredConstraint(
                    column=constraint.fq_column,
                    placeholder=constraint.placeholder,
                )

        # verify that the query is constrained by at least one of the
        # requester's identities. this means that the results of the query will
        # be restricted to data that only the requester has access to.
        #
        # we look at both the identities returned explicitly from the
        # `requester_identities` method, and also the identities that have been
        # annotated in the join conditions.
        req_ids = self.requester_identities()
        join_identities = chain.from_iterable(
            j.requester_identities for j in self.allowed_joins()
        )
        all_req_identities = set(chain(req_ids, join_identities))
        if all_req_identities:
            found_id = False
            for one_id in all_req_identities:
                if one_id in facets.required_constraints:
                    found_id = True
                    break
            if not found_id:
                raise exc.MissingRequiredIdentity(identities=all_req_identities)

        # check that the query limits the rows correctly, if we restrict to a limit
        if (limit := self.max_limit()) is not None:
            if facets.limit > limit:
                raise exc.TooManyRows(limit=facets.limit)

        # check that every function used has been allowlisted
        for fn in facets.functions:
            if not self.can_use_function(fn):
                raise exc.IllegalFunction(function=fn)



from typing import Optional

import lark

from heimdallm.support.github import make_ambiguous_parse_issue

from .utils import FqColumn, JoinCondition, RequiredConstraint


class BaseException(Exception):
    """a convenience class for all heimdallm sql exceptions, to make all of them
    easier to catch"""


class GeneralParseError(BaseException):
    """thrown from our grammar parser for anything that we want to bubble up as an
    InvalidQuery exception, but we don't have access to the raw input from within the
    parser"""


class InvalidQuery(BaseException):
    """thrown when our grammar cannot parse the LLM output"""

    def __init__(self, *, query: str):
        super().__init__(f"\n\n{query}\n")
        self.query = query


class ReservedKeyword(BaseException):
    def __init__(self, *, keyword: str):
        super().__init__(f"Alias `{keyword}` is a reserved keyword")
        self.keyword = keyword


class AmbiguousParse(BaseException):
    def __init__(self, *, trees: list[lark.ParseTree], query: str):
        self.issue_link = make_ambiguous_parse_issue(query, trees)

        super().__init__(
            f"Query resulted in {len(trees)} ambiguous parse trees. "
            "Please report this query to the HeimdaLLM maintainer using "
            f"the following link:\n{self.issue_link}"
        )
        self.trees = trees
        self.query = query


class UnqualifiedColumn(BaseException):
    """thrown when a column isn't fully qualified with the table prefix"""

    def __init__(self, column: Optional[str] = None):
        message = "Fully-qualified column name needs to be in the form 'table.column'"
        if column is not None:
            message += f" (got {column!r})"

        super().__init__(message)
        self.column = column


class IllegalSelectedColumn(BaseException):
    def __init__(self, *, column: str):
        """
        :param column: The column that was selected. This is not a FqColumn because we
        may not always have a table name.
        """
        message = f"Column `{column}` is not allowed in SELECT"
        super().__init__(message)
        self.column = column


class IllegalConditionColumn(BaseException):
    def __init__(self, *, column: FqColumn):
        message = f"Column `{column}` is not allowed in WHERE"
        super().__init__(message)
        self.column = column


class MissingRequiredConstraint(BaseException):
    def __init__(self, *, column: FqColumn, placeholder: str):
        message = f"Missing required constraint `{column}`=:{placeholder}"
        super().__init__(message)
        self.column = column
        self.placeholder = placeholder


class MissingRequiredIdentity(BaseException):
    def __init__(self, *, identities: set[RequiredConstraint]):
        message = f"Missing one required identities: {identities!r}"
        super().__init__(message)
        self.identities = identities


class IllegalJoinCondition(BaseException):
    def __init__(self, message, *, table, condition):
        super().__init__(message)
        self.table = table
        self.condition = condition


class IllegalJoinTable(BaseException):
    def __init__(self, *, join: JoinCondition):
        message = f"Join condition {join} is not allowed"
        super().__init__(message)
        self.join = join


class IllegalJoinType(BaseException):
    def __init__(self, *, join_type: str):
        message = f"JOIN type `{join_type}` is not allowed"
        super().__init__(message)
        self.join_type = join_type


class DisconnectedTable(BaseException):
    def __init__(self, *, table):
        message = f"Table `{table}` is not connected to the query"
        super().__init__(message)
        self.table = table


class BogusJoinedTable(BaseException):
    def __init__(self, *, table):
        message = f"Join condition for `{table}` does not reference the table"
        super().__init__(message)
        self.table = table


class TooManyRows(BaseException):
    def __init__(self, *, limit):
        message = f"Attempting to return too many rows ({limit})"
        super().__init__(message)
        self.limit = limit


class IllegalFunction(BaseException):
    def __init__(self, *, function):
        message = f"Function `{function}` is not allowed"
        super().__init__(message)
        self.function = function



from typing import Any, Optional, Sequence

from . import exc


class RequiredConstraint:
    """represents a constraint that must be applied to the query. in the query,
    this comes in the form of "table.column=:placeholder". enforced by the
    grammar, the comparison is always equality, the left hand side is always a
    fully-qualified column, and the right hand side is always a placeholder.
    this requirements ensure that the query is always constrained by a value
    that the developer specifies at query execution time."""

    def __init__(self, *, column: str, placeholder: str):
        self.fq_column = FqColumn.from_string(column)
        self.placeholder = placeholder

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, RequiredConstraint)
            and other.fq_column == self.fq_column
            and other.placeholder == self.placeholder
        )

    def __hash__(self):
        return hash((self.fq_column, self.placeholder))

    def __str__(self):
        return f"{self.fq_column}=:{self.placeholder}"

    __repr__ = __str__


class FqColumn:
    """represents a fully-qualified column name. we require that LLM-produced
    queries use fully-qualified columns in their SELECT and WHERE clauses,
    otherwise we would need to infer which table owned the column, which
    requires runtime database analysis. much easier to require the LLM give us
    fully-qualified names"""

    __slots__ = ("table", "column")

    @classmethod
    def from_string(cls, s: str):
        if "." not in s:
            raise exc.UnqualifiedColumn(s)
        table, column = s.split(".")
        return cls(table=table, column=column)

    def __init__(self, *, table: str, column: str):
        self.table = table
        self.column = column

    def __str__(self):
        return f"{self.table}.{self.column}"

    name = property(str)

    def __iter__(self):
        return iter((self.table, self.column))

    def __hash__(self):
        return hash((self.table, self.column))

    def __eq__(self, other):
        return self.table == other.table and self.column == other.column

    def __repr__(self):
        return f"{self.table}.{self.column}"


class JoinCondition:
    """represents an equi-join between two tables. for our hash and equality
    functions, we don't care about the order of the join condition"""

    __slots__ = ("first", "second", "identity_placeholder")

    def __init__(self, first: str, second: str, *, identity: Optional[str] = None):
        """if identity is True, then either column in the join condition is a
        valid requester identity. this means it can be used to constrain the
        query in the same way SQLConstraintValidator.requester_identities
        does."""
        self.first = FqColumn.from_string(first)
        self.second = FqColumn.from_string(second)
        self.identity_placeholder = identity

    @property
    def requester_identities(self) -> Sequence[RequiredConstraint]:
        """if this join condition has been marked as an identity join,
        construct the required constraints for both sides of the join. we'll use
        those constraints when testing for the requester's identity"""
        if self.identity_placeholder:
            return [
                RequiredConstraint(
                    column=self.first.name,
                    placeholder=self.identity_placeholder,
                ),
                RequiredConstraint(
                    column=self.second.name,
                    placeholder=self.identity_placeholder,
                ),
            ]
        else:
            return []

    def __hash__(self):
        # it shouldn't matter which order the join condition comes in, the hash
        # should be the same
        return hash((self.first, self.second)) ^ hash((self.second, self.first))

    def __eq__(self, other):
        return (self.first == other.first and self.second == other.second) or (
            self.first == other.second and self.second == other.first
        )

    def __str__(self) -> str:
        return f"{self.first}={self.second}"

    __repr__ = __str__


class _AnyJoinCondition(JoinCondition):
    """a convenience class for representing any valid join condition."""

    def __init__(self):
        super().__init__("*.*", "*.*")


ANY_JOIN = _AnyJoinCondition()



from enum import Enum
from typing import Callable

import openai

from heimdallm import llm


class OpenAIMethod(Enum):
    """How the OpenAI API should complete the prompt. OpenAI provides two main methods
    of completion, both appear to be similar in quality. TODO document the differences.
    """

    CHAT = "chat"
    COMPLETION = "completion"


class Client(llm.LLMIntegration):
    """The LLM integration with OpenAI's ChatGPT and Completion API.

    :param api_key: Your secret OpenAI API key.
    :param model: The model to use. Prefer a higher model with a high context window, as
        the prompt envelopes can make the LLM input quite long.
    :param method: The method to use to interact with the OpenAI API. I don't really
        know the difference in terms of quality, but the chat method has worked fine so
        far."""

    def __init__(
        self,
        api_key: str,
        model: str = "gpt-3.5-turbo-16k",
        method: OpenAIMethod = OpenAIMethod.CHAT,
    ):
        self.api_key = api_key
        self.model = model
        self.method = method

    def complete(self, untrusted_user_input: str) -> str:
        """Complete the untrusted user input and return some structured output.

        :param untrusted_user_input: The untrusted user input.
        :return: The structured output."""
        fn_map: dict[OpenAIMethod, Callable[[str], str]] = {
            OpenAIMethod.CHAT: self._complete_via_chatgpt,
            OpenAIMethod.COMPLETION: self._complete_via_completion,
        }
        fn = fn_map[self.method]
        return fn(untrusted_user_input)

    def _complete_via_chatgpt(self, untrusted_user_input: str) -> str:
        chat_completion = openai.ChatCompletion.create(
            api_key=self.api_key,
            model=self.model,
            messages=[
                {
                    "role": "user",
                    "content": untrusted_user_input,
                }
            ],
        )
        untrusted_llm_output = chat_completion.choices[0].message.content
        return untrusted_llm_output

    def _complete_via_completion(self, prompt):
        response = openai.Completion.create(
            engine=self.model,
            prompt=prompt,
            max_tokens=256,
            n=1,
            stop=None,
            temperature=0.7,
        )
        return response.choices[0].text.strip()



import sqlite3
from pathlib import Path
from typing import Callable, Sequence, Union

import lark
from lark import Lark, ParseTree
from lark.exceptions import VisitError

from heimdallm.bifrost import Bifrost as _Bifrost
from heimdallm.bifrosts.sql import exc
from heimdallm.envelope import PromptEnvelope as _PromptEnvelope
from heimdallm.llm import LLMIntegration
from heimdallm.llm_providers.mock import EchoMockLLM

from .envelope import TestSQLPromptEnvelope
from .validator import SQLConstraintValidator
from .visitors import AmbiguityResolver

_THIS_DIR = Path(__file__).parent
_GRAMMAR_PATH = _THIS_DIR / "sqlite.lark"


def _build_grammar() -> Lark:
    """
    returns a limited sqlite SELECT grammar

    noteworthy:
        - no outer joins
        - no subqueries

    theoretically, subqueries could be allowed, but it would be more work, and
    i'm not yet convinced that an LLM produces subqueries often enough to make
    it worth it.

    outer joins are unsafe because the join constraint is not applied to the
    rows that would be considered "outer."
    """
    with open(_GRAMMAR_PATH, "r") as h:
        grammar = Lark(
            ambiguity="explicit",
            maybe_placeholders=False,
            grammar=h,
        )
    return grammar


def _build_tree_producer() -> Callable[[Lark, str], ParseTree]:
    def parse(grammar: Lark, untrusted_query: str) -> ParseTree:
        ambig_tree = grammar.parse(untrusted_query)
        try:
            final_tree = AmbiguityResolver(untrusted_query).transform(ambig_tree)
        except VisitError as e:
            if isinstance(e.orig_exc, exc.BaseException):
                raise e.orig_exc
            raise e
        return final_tree

    return parse


def get_schema(conn: sqlite3.Connection):
    """a convenience function to get the schema of a sqlite database. you
    probably want to write your own function to do this, one that doesn't
    include tables and columns that you care about sending to the LLM"""
    schema = []
    for line in conn.iterdump():
        if "CREATE TABLE" in line:
            schema.append(line)
    return "\n".join(schema)


class SQLBifrost(_Bifrost):
    """A Bifrost for traversing SQL ``SELECT`` queries."""

    # for tests
    @classmethod
    def mocked(
        cls,
        constraint_validators: Union[
            SQLConstraintValidator, Sequence[SQLConstraintValidator]
        ],
    ):
        """A convenience method for our tests. This creates a Bifrost that assumes its
        untrusted input is a SQL query already, so it does not need to communicate with
        the LLM, only parse and validate it.

        :param constraint_validators: A constraint validator or sequence of constraint
            validators to run on the untrusted input.
        """
        if not isinstance(constraint_validators, Sequence):
            constraint_validators = [constraint_validators]

        llm = EchoMockLLM()
        return cls(
            llm=llm,
            constraint_validators=constraint_validators,
            prompt_envelope=TestSQLPromptEnvelope(
                llm=llm,
                db_schema="<schema>",  # doesn't matter
                validators=constraint_validators,
            ),
        )

    def __init__(
        self,
        *,
        llm: LLMIntegration,
        prompt_envelope: _PromptEnvelope,
        constraint_validators: Sequence[SQLConstraintValidator],
    ):
        super().__init__(
            llm=llm,
            prompt_envelope=prompt_envelope,
            grammar=_build_grammar(),
            tree_producer=_build_tree_producer(),
            constraint_validators=constraint_validators,
        )

    def parse(self, untrusted_llm_output: str) -> ParseTree:
        """Parse the unwrapped SQL query from the LLM's output. Raise a SQL-specific
        exception if the query is not valid.

        :param untrusted_llm_output: The output from the LLM, which should be a SQL
            query. If it isn't, then our :meth:`SQLPromptEnvelope.unwrap` method failed.
        """
        try:
            return super().parse(untrusted_llm_output)
        except lark.exceptions.UnexpectedEOF as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e
        except lark.exceptions.UnexpectedCharacters as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e
        except exc.BaseException as e:
            raise e
        except Exception as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e



from abc import ABC, abstractmethod

from lark import Lark, ParseTree


class ConstraintValidator(ABC):
    @abstractmethod
    def fix(self, Grammar: Lark, tree: ParseTree) -> str:
        """if the tree does not pass the constraints, attempt to reconstruct it
        to satisfy the constraints. if the tree cannot be reconstructed, throw a
        bifrost-specific exception"""
        pass

    @abstractmethod
    def validate(self, untrusted_input: str, tree: ParseTree):
        """throws a bifrost-specific exception if the tree does not pass the
        constraints. the untrusted_input is provided for context if we need to raise
        an exception that contains the original input."""
        pass



safe_date_functions = {
    "adddate",
    "addtime",
    "convert_tz",
    "curdate",
    "current_date",
    "current_time",
    "current_timestamp",
    "curtime",
    "date",
    "date_add",
    "date_format",
    "date_sub",
    "datediff",
    "day",
    "dayname",
    "dayofmonth",
    "dayofweek",
    "dayofyear",
    "extract",
    "from_days",
    "from_unixtime",
    "get_format",
    "hour",
    "last_day",
    "localtime",
    "localtimestamp",
    "makedate",
    "maketime",
    "microsecond",
    "minute",
    "month",
    "monthname",
    "now",
    "period_add",
    "period_diff",
    "quarter",
    "sec_to_time",
    "second",
    "str_to_date",
    "subdate",
    "subtime",
    "sysdate",
    "time",
    "time_format",
    "time_to_sec",
    "timediff",
    "timestamp",
    "timestampadd",
    "timestampdiff",
    "to_days",
    "to_seconds",
    "unix_timestamp",
    "utc_date",
    "utc_time",
    "utc_timestamp",
    "week",
    "weekday",
    "weekofyear",
    "year",
    "yearweek",
}

safe_string_functions = {
    "ascii",
    "bin",
    "bit_length",
    "char",
    "char_length",
    "character_length",
    "concat",
    "concat_ws" "elt",
    "export_set",
    "field",
    "find_in_set",
    "format",
    "from_base64",
    "hex",
    "insert",
    "instr",
    "lcase",
    "left",
    "length",
    # this is not safe
    # "load_file",
    "locate",
    "lower",
    "lpad",
    "ltrim",
    "make_set",
    "mid",
    "oct",
    "octet_length",
    "ord",
    "position",
    "quote",
    "regexp_instr",
    "regexp_like",
    "regexp_replace",
    "regexp_substr",
    "repeat",
    "replace",
    "reverse",
    "right",
    "rpad",
    "rtrim",
    "soundex",
    "space",
    "strcmp",
    "substr",
    "substring",
    "substring_index",
    "to_base64",
    "trim",
    "ucase",
    "unhex",
    "upper",
    "weight_string",
}

safe_math_functions = {
    "abs",
    "acos",
    "asin",
    "atan",
    "atan2",
    "ceil",
    "ceiling",
    "conv",
    "cos",
    "cot",
    "crc32",
    "degrees",
    "div",
    "exp",
    "floor",
    "ln",
    "log",
    "log10",
    "log2",
    "mod",
    "pi",
    "pow",
    "power",
    "radians",
    "rand",
    "round",
    "sign",
    "sin",
    "sqrt",
    "tan",
    "truncate",
}


safe_misc_functions = {
    "bit_count",
    "if",
    "ifnull",
    "nullif",
}

safe_json_functions = {
    "json_array",
    "json_array_append",
    "json_array_insert",
    "json_contains",
    "json_contains_path",
    "json_depth",
    "json_extract",
    "json_insert",
    "json_keys",
    "json_length",
    "json_merge",
    "json_merge_patch",
    "json_merge_preserve",
    "json_object",
    "json_overlaps",
    "json_pretty",
    "json_quote",
    "json_remove",
    "json_replace",
    "json_schema_valid",
    "json_schema_validation_report",
    "json_search",
    "json_set",
    "json_type",
    "json_unquote",
    "json_valid",
    "json_value",
}

safe_agg_functions = {
    "avg",
    "bit_and",
    "bit_or",
    "bit_xor",
    "count",
    "group_concat",
    "json_arrayagg",
    "json_objectagg",
    "max",
    "min",
    "std",
    "stddev",
    "stddev_pop",
    "stddev_samp",
    "sum",
    "var_pop",
    "var_samp",
    "variance",
}

safe_functions = (
    safe_date_functions
    | safe_string_functions
    | safe_math_functions
    | safe_misc_functions
    | safe_json_functions
    | safe_agg_functions
)



import math
from collections import defaultdict as dd
from typing import MutableMapping, Optional

from lark import Token, Transformer, Tree, Visitor

from ... import exc
from ...utils import FqColumn, JoinCondition, RequiredConstraint
from ..utils.identifier import get_identifier
from ..utils.visitors import AliasCollector


class Facets:
    """this simple class is used to collect all of the facets of a query, so
    that we can easily validate them with a constraint validator"""

    def __init__(self) -> None:
        # tables that were joined but do not reference the table in the join
        # condition. if this list has values, validation fails.
        self.bad_joins: list[str] = []
        # the table selected in the FROM clause. this will always be set after a
        # successful parse
        self.selected_table: Optional[str] = None
        # the tables joined to the query, this could include the table selected
        # in the FROM clause, if there are JOINs in the query
        self.joined_tables: MutableMapping[str, set[JoinCondition]] = dd(set)
        # the columns selected in the query
        self.selected_columns: set[FqColumn] = set()
        # the columns used in the WHERE, JOIN, HAVING, and ORDER BY clauses
        self.condition_columns: set[FqColumn] = set()
        # the required conditions in the WHERE and JOIN clauses, used to
        # constrain the query so that it is safe
        self.required_constraints: set[RequiredConstraint] = set()
        # all of the functions used in the query
        self.functions: set[str] = set()
        # the row limit of the query
        self.limit = math.inf


class FacetCollector(Visitor):
    """collects all of the facets of the query that we care about. this will
    feed directly into the constraint validator"""

    def __init__(self, facets: Facets, collector: AliasCollector):
        self._collector = collector
        self._facets = facets

    def _resolve_column(self, node):
        if node.data == "column_alias":
            maybe_alias = get_identifier(node)
            table, column = self._collector._aliased_columns.get(
                maybe_alias, (None, maybe_alias)
            )
            return table, column

        # should never happen
        raise RuntimeError(f"unknown column reference type: {type(node)}")

    def _resolve_table(self, table_ref):
        if isinstance(table_ref, Tree):
            if table_ref.data == "aliased_table":
                table_name = get_identifier(table_ref)
                return table_name

            elif table_ref.data == "table_name":
                # we have no way of knowing if this is an alias or not, other
                # than testing for its existence in the collector's alias map
                maybe_alias = get_identifier(table_ref)
                table = self._collector._aliased_tables.get(maybe_alias, maybe_alias)
                return table

        elif isinstance(table_ref, Token):
            return table_ref.value

        # should never happen
        raise RuntimeError(f"unknown table reference type: {type(table_ref)}")

    def join(self, node):
        if join_type := list(node.find_data("illegal_join")):
            join_type = join_type[0].children[0].type
            raise exc.IllegalJoinType(join_type=join_type)

        joined_table = node.children[1].children[0]
        joined_table_name = self._resolve_table(joined_table)

        # if a required_comparison node exists, it means it is actually required
        # (enforced by the grammar, see grammar comments). a required comparison has
        # a placeholder for the RHS
        for required_comparison in node.find_data("required_comparison"):
            self._add_required_comparison(required_comparison)

        for condition in node.find_data("connecting_join_condition"):
            # from_table may be an alias, but from_column will always be authoritative.
            # the LHS of the join condition is always a fully-qualified column
            from_fq_column_node = condition.children[0]
            from_table_node, from_column_node = from_fq_column_node.children
            from_column = get_identifier(from_column_node)
            from_table = self._resolve_table(from_table_node)

            # conditions in a join are compared against are allowed conditions, so
            # record the LHS
            self._collect_condition_column(from_fq_column_node)

            # to_table may be an alias, but to_column will always be authoritative.
            # the RHS of the join condition can be a `value` rule.
            to_fq_column_node = condition.children[2]

            # token? it's probably a string or number, keep going
            if isinstance(to_fq_column_node, Token):
                continue
            # not a fq column? it's probably a function or something nested, keep going.
            elif (
                isinstance(to_fq_column_node, Tree)
                and to_fq_column_node.data != "fq_column"
            ):
                continue

            to_table, to_column_node = to_fq_column_node.children
            to_column = get_identifier(to_column_node)

            to_table = self._resolve_table(to_table)

            # the joined table must be one of the parts of the join condition
            if joined_table_name != from_table and joined_table_name != to_table:
                self._facets.bad_joins.append(joined_table_name)
                continue

            # conditions in a join are compared against are allowed conditions, so
            # record the RHS
            self._collect_condition_column(to_fq_column_node)

            # our join represents two sides, the from table and the to table.
            # we'll record both in our joined_tables set.
            join_spec = JoinCondition(
                f"{from_table}.{from_column}",
                f"{to_table}.{to_column}",
            )
            self._facets.joined_tables[from_table].add(join_spec)
            self._facets.joined_tables[to_table].add(join_spec)

    def selected_table(self, node):
        table_node = list(node.find_data("table_name"))[0]
        table_name = get_identifier(table_node)
        self._facets.selected_table = table_name

    def selected_column(self, node):
        child = node.children[0]

        if isinstance(child, Token) and child.type == "COUNT_STAR":
            return

        elif isinstance(child, Token) and child.type == "ALL_COLUMNS":
            raise exc.IllegalSelectedColumn(column="*")

        elif isinstance(child, Tree):
            # if it's an aliased column, we need to ensure that the thing being aliased
            # isn't a non-fully-qualified column. we do that by looking for
            # `column_alias`
            if child.data == "aliased_column":
                alias_child = child.children[0]
                # if we're aliasing COUNT(*), it's safe to ignore, since it doesn't
                # reveal any of the underlying values
                if isinstance(alias_child, Token) and alias_child.type == "COUNT_STAR":
                    return

                if list(alias_child.find_data("column_alias")):
                    alias = get_identifier(alias_child)
                    raise exc.UnqualifiedColumn(column=alias)

            # if the column looks like a column alias (meaning a non-fully-qualified
            # column), then that's an error, because we only work with fully-qualified
            # columns
            elif child.data == "column_alias":
                alias = get_identifier(child)
                raise exc.UnqualifiedColumn(column=alias)

            # if we're not an aliased column, but we contain a column alias somewhere,
            # that's an error, because we only work with fully-qualified columns
            elif column_alias := list(child.find_data("column_alias")):
                alias = get_identifier(column_alias[0])
                raise exc.UnqualifiedColumn(column=alias)

            # if we've made it this far, we're sure we're dealing with a fully-qualified
            # column, or a non-column based expression
            try:
                table_node = next(child.find_data("table_name"))
            # it's some non-column expression, which we don't care about
            except StopIteration:
                pass
            # there's a fully qualified column there, so we'll record it
            else:
                table_name = self._resolve_table(table_node)
                # column_name will always be authoritative, even if it is aliased in
                # this node
                column_name = get_identifier(list(child.find_data("column_name"))[0])
                self._facets.selected_columns.add(
                    FqColumn(
                        table=table_name,
                        column=column_name,
                    )
                )

    def _add_required_comparison(self, node):
        """takes a node representing a required comparison and adds it to the
        facets"""
        maybe_fq_column_node, placeholder = node.children
        placeholder_name = placeholder.children[0].value

        if maybe_fq_column_node.data == "column_alias":
            table_name, column_name = self._resolve_column(maybe_fq_column_node)

        elif maybe_fq_column_node.data == "fq_column":
            fq_column_node = maybe_fq_column_node
            table_node, column_node = fq_column_node.children

            table_name = self._resolve_table(table_node)
            column_name = get_identifier(column_node)
        else:
            raise RuntimeError(
                f"Unknown required column type {type(maybe_fq_column_node)}"
            )

        self._facets.required_constraints.add(
            RequiredConstraint(
                column=f"{table_name}.{column_name}",
                placeholder=placeholder_name,
            )
        )

    def where_clause(self, where_node: Tree):
        """here we'll do a breadth-first search on the where clause, going level
        by level. if any level contains an "OR", that means the entire level is
        tainted and can't be used for a required constraint, because the
        required constraint may be optional.

        if a level is tainted, then all of its children (WHERE subclauses) are
        tainted as well, so we will not process them.

        the end result is that we only collect a required comparison node IFF it
        is not joined by an OR anywhere in the WHERE clause, either at its
        level, or at any level above it. only then can we know that the
        comparison is actually constraining to the query"""
        conditions = where_node.children[1]
        stack: list[Tree] = [conditions]

        while stack:
            conditions = stack.pop()

            level_stack: list[Tree] = []
            for child in conditions.children:
                if isinstance(child, Token):
                    # is the level tainted? if so, clear the stack and exit the loop
                    if child.type == "WHERE_TYPE" and child.value.lower() == "or":
                        level_stack = []
                        break
                elif isinstance(child, Tree):
                    level_stack.append(child)

            for child in reversed(level_stack):
                stack.append(child)
                if child.data == "required_comparison":
                    self._add_required_comparison(child)

    def _collect_condition_column(self, node):
        """here we'll parse out the columns that are referenced anywhere in the
        WHERE, regardless of the depth of the expression. we care if a column is being
        referenced at all, even optionally, because that will be checked against the
        allowlist"""
        for fq_column_node in node.find_data("fq_column"):
            table_node, column_node = fq_column_node.children
            table_name = self._resolve_table(table_node)
            column_name = get_identifier(column_node)
            self._facets.condition_columns.add(
                FqColumn(
                    table=table_name,
                    column=column_name,
                )
            )

        for column_alias_node in node.find_data("column_alias"):
            table_name, column_name = self._resolve_column(column_alias_node)
            if not table_name:
                # no table name AND no column name means the inserted alias is an alias
                # for a non-column expression, like `COUNT(*) as thing`. this is valid
                # as it is, so we don't raise, and we don't track it as a condition
                # column.
                if not column_name:
                    pass
                else:
                    raise exc.UnqualifiedColumn(column=column_name)
            self._facets.condition_columns.add(
                FqColumn(
                    table=table_name,
                    column=column_name,
                )
            )

    where_condition = _collect_condition_column
    having_condition = _collect_condition_column
    order_column = _collect_condition_column

    def limit(self, node):
        self._facets.limit = float(node.children[0].value)

    def function_name(self, node):
        self._facets.functions.add(node.children[0].value.lower())


class AmbiguityResolver(Transformer):
    """this transformer's purpose is to resolve ambiguities in the parse tree
    that can only be resolved through some extra knowledge that would be
    difficult to embed in the grammar itself.
    """

    def __init__(self, query: str):
        self.query = query
        super().__init__()

    def test_alias(self, i, tree, trees) -> bool:
        """
        This resolves ambiguities in the parse tree related to aliases. For
        example, the following query is ambiguous:

            select t1.secret from t1 left join t2 on t1.jid = t2.jid

        Is "left" an alias for "t1", or is it part of the "left join"? It's
        ambiguous. We know it's not ambiguous because "left" is a keyword, but
        the parser can't know this. This class resolves those ambiguities by
        looking at all possible ambiguous parse trees in the ambiguity, and
        selecting only the one that does not have an alias conflict with a
        reserved keyword.
        """
        for alias_node in tree.find_data("generic_alias"):
            try:
                get_identifier(alias_node)
            except exc.ReservedKeyword:
                return False
        return True

    def test_req_comparisons(self, i, tree, trees) -> bool:
        """in a where_condition rule, the children can include
        `relational_comparison` and `required_comparison` rules, which are
        ambiguous, because required comparisons are a subset of relational
        comparisons. we always prefer to interpret the ambiguity as a required
        comparison though, because it is more strict, and it satisfies our
        required comparison validator constraints"""
        if tree.data in ("where_condition", "join_condition"):
            return tree.children[0].data == "required_comparison"
        return True

    def test_arith_expr(self, i, tree, trees) -> bool:
        """the arith_expr node is recursive, so it can be ambiguous. just choose the
        first parse of it, since this node doesn't really matter"""
        if tree.data == "arith_expr":
            return i == 0
        return True

    def _ambig(self, trees):
        def test_tree(i, tree):
            return (
                self.test_alias(i, tree, trees)
                and self.test_req_comparisons(i, tree, trees)
                and self.test_arith_expr(i, tree, trees)
            )

        pruned_trees = [tree for i, tree in enumerate(trees) if test_tree(i, tree)]

        if len(pruned_trees) == 0:
            raise exc.InvalidQuery(query=self.query)
        elif len(pruned_trees) == 1:
            return pruned_trees[0]
        else:
            raise exc.AmbiguousParse(trees=pruned_trees, query=self.query)
